/* Modifierad suffix.c som fixar till SLlist fr Granskas behov. 
   Copyright (C) 1998-2001
   Viggo Kann
   viggo@nada.kth.se
*/

/******************************************************************************

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

******************************************************************************/

#include "stava.h"
#include "stavaconstants.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef struct suffixchecklist {
  unsigned char *checksuffix;
  int suffixlen;
  struct suffixchecklist *next;
  int accept;
} suffixchecklist;

#ifndef MAXNOOFFORBIDDENENDINGS
#define MAXNOOFFORBIDDENENDINGS 6
#endif
#define MAXLINELEN 1000

typedef struct suffixset {
  /* Ingngssuffix (framlnges) och dess lngd: */
  unsigned char *suffix;
  int len;
  /* Lista med uppslagssuffix: */
  struct suffixchecklist *check;
  /* Binrt sktrd sorterat efter reversesuffix: */
  struct suffixset *left, *right; 
  /* Positiv eller negativ regel: */
  int accept;
  /* Tilltna och otilltna tecken i pos precis fre ndelsen: */
  unsigned char *allowedlastcharacters; 
  unsigned char *forbiddenlastcharacters;
  /* Otilltna strngar precis fre ndelsen: */
  unsigned char *forbiddenending[MAXNOOFFORBIDDENENDINGS + 1];
  int forbiddenendinglength[MAXNOOFFORBIDDENENDINGS];
  /* Ska skningen sluta efter frsta stavelsen eller vid ordets brjan: */
  int stopsearchatbeginning;
} suffixset;

static int linecount = 0; /* radnummer i aktuellt bjningsmnster */
static char *kommentar = NULL; /* senast inlsta kommentar */
static char wordclass[100];

static void InitieraBokstavsTabeller(void)
{ int i;
  /* fyll i 0-flt med standardvrde i bokstavstabellen till ISO */
  for (i = 0; i < 256; i++) {
    if (intern_ISO[i] == 0) intern_ISO[i] = i;
  }
}

/* WriteISO skriver ut en ASCII-textstrng versatt till ISO Latin-1 */
void WriteISO(const unsigned char *s)
{
  while (*s) putchar(intern_ISO[(unsigned char)*s++]);
}

void WriteWithoutTags(unsigned char *s)
{
  for (; *s; s++) {
    if (*s == '<') {
      while (*s && *s != '>') s++;
      if (!*s) return; /* oavslutad tagg */
    } else putchar(*s);
  }
}

/* NextWord letar efter nsta ord i s och stter start att peka p brjan av
 * det. Efter ordet stts '\0' in och efterfljande position returneras.
 * Om inget ord finns i s returneras NULL.
 * @ betyder tomma strngen. # gr resten av raden till kommentar */
static unsigned char *NextWord(unsigned char *s, unsigned char **start)
{
  while (*s && *s < '@' && *s != '#' && *s != '-') s++;
  if (*s == '\0' || *s == '#') {
    kommentar = s;
    return NULL;
  }
  *start = s;
  if (*s != '@')
    while (*s >= 'A' || *s == '-') {
      if (*s == '~') *s = '_'; else *s = ISO_intern[*s];
      s++;
    }
  if (*s == '\0') return s;
  if (*s == '#') s[1] = '#';
  *s = '\0';
  return s + 1;
}

/* CopyTag letar upp nsta < i t, kopierar allt mellan < och > till
tag och returnerar en pekare till positionen efter >.
Om det inte finns ngot par < > i t s stts *tag till 0 och NULL returneras */
unsigned char *CopyTag(unsigned char *tag, unsigned char *t)
{ unsigned char *s, *tagp;
  *tag = '\0';
  if (!t) return NULL;
  while (*t && *t != '<') t++;
  if (!*t) return NULL;
  for (s = t + 1, tagp = tag; *s && *s != '>'; s++, tagp++) *tagp = *s;
  if (!*s) { *tag = '\0'; return NULL; }
  *tagp = '\0';
  return s + 1;
}

/* IsSuffix kollar om t r ett suffix till s */
static int IsSuffix(char *s, char *t)
{ int si = strlen(s), ti = strlen(t);
  if (si < ti) return 0;
  s += si - 1;
  for (t += ti - 1; ti > 0; ti--, s--, t--)
    if (*s != *t) return 0;
  return 1;
}


/* WriteSuffix skriver ut en regelrad med taggmrkning. Om *tag != 0
s skrivs den taggen ut. Annars rknar proceduren ut vilken taggen ska
vara, skriver ut den och kopierar den till tag. */
static void WriteSuffix(suffixset *suf, char *suffix, int tagtype,
			char *tag)
{ suffixchecklist *p;
  unsigned char **fep = suf->forbiddenending;
  char *t;
  char lastchar = '\0';
  if (*suffix) lastchar = suffix[strlen(suffix)-1];

  if (*tag) ; else
  if (strcmp(wordclass, "nn") == 0) {
    p = suf->check;
    if (p && p->accept) {
      char *s = suffix, *t = p->checksuffix;
      if (*s == 'm' && *t == 'm') {
	while (*s == 'm') s++;
	while (*t == 'm') t++;
      }
      while (*s && *s == *t) { s++; t++; }
      if (strncmp(s, "ing", 3) == 0) {
	tagtype = 'n';
	if (strcmp(s, "ing") == 0) linecount = 1; }
      else if (strcmp(s, "erska") == 0 || strcmp(s, "inna") == 0) {
	tagtype = 'n';
	linecount = 1; }
      else if (strcmp(s, "et") == 0) {
	linecount = 2;
	if (tagtype == 0) tagtype = 't'; }
      else if (strcmp(s, "er") == 0) {
	linecount = 3;
	if (tagtype == 0) tagtype = 'n'; }
      else if (IsSuffix(s, "ns") && tagtype != 't') {
	linecount = 5; tagtype = 'n'; }
      else if (IsSuffix(s, "ts")) { linecount = 5; tagtype = 't'; }
      else if (IsSuffix(s, "rs")) { 
	linecount = 6; 
	if (tagtype == 0) tagtype = 'n'; }
      else if (IsSuffix(s, "rna") || strcmp(s, "na") == 0) {
	linecount = 7;
	if (tagtype == 0) tagtype = 'n'; }
      else if (IsSuffix(s, "rnas") || strcmp(s, "nas") == 0) {
	linecount = 8;
	if (tagtype == 0) tagtype = 'n'; }
      else if (IsSuffix(suffix, "aren")) { linecount = 2; tagtype = 'n'; }
      else if (tagtype == 't') {
	if (strcmp(s, "en") == 0 || strcmp(s, "len") == 0
	    || strcmp(s, "ren") == 0 || strcmp(s, "nen") == 0 
	    || strcmp(s, "onen") == 0) linecount = 7;
	else if (strcmp(s, "ens") == 0 || strcmp(s, "lens") == 0
	    || strcmp(s, "rens") == 0 || strcmp(s, "nens") == 0) linecount = 8;
      }
      else if (tagtype == 0 || tagtype == 'n') {
	int oldtagtype = tagtype;
	tagtype = 'n';
	if (*s == '\0') linecount = 1;
	else if (strcmp(s, "n") == 0 || strcmp(s, "en") == 0) linecount = 2;
	else if (strcmp(s, "or") == 0 || strcmp(s, "ar") == 0
		 || strcmp(s, "rar") == 0
		 || strcmp(s, "er") == 0 || strcmp(s, "r") == 0) linecount = 3;
	else if (strcmp(s, "s") == 0) linecount = 4;
	else tagtype = oldtagtype;
      }
    }
  } else
  if (strcmp(wordclass, "jj") == 0) {
    p = suf->check;
    if (p && p->accept) {
      char *s = suffix, *t = p->checksuffix;
      int moved = 0;
      if (t[0] == 'e' && t[1] == *s && t[2] == '\0') { /* el/en */
	s++; t += 2;
      }
      if (*s == 'm' && *t == 'm') {
	while (*s == 'm') s++;
	while (*t == 'm') t++;
      }
      while (*s && *s == *t) { s++; t++; moved = 1; }
      if (moved && s[-1] == 'm' && t[-1] == 'm') {
	while (*s == 'm') s++;
	while (*t == 'm') t++;
      }

      if (IsSuffix(suffix, "het")) { 
	linecount = 1; tagtype = 'n'; }
      else if (IsSuffix(suffix, "heten")) { 
	linecount = 2; tagtype = 'n'; }
      else if (strncmp(s, "het", 3) == 0 || strncmp(suffix, "het", 3) == 0) {
	tagtype = 'n'; }
      else if (IsSuffix(suffix, "ast") || IsSuffix(suffix, "erst")) 
	tagtype = 1009; /* jj.suv.utr/neu.sin/plu.ind.nom */
      else if (IsSuffix(suffix, "aste") || IsSuffix(suffix, "ersta")) 
	tagtype = 1010; /* jj.suv.utr/neu.sin/plu.def.nom */
      else if (IsSuffix(suffix, "astes") || IsSuffix(suffix, "erstas")) 
	tagtype = 1011; /* jj.suv.utr/neu.sin/plu.def.gen */
      else if (*s == '\0') {
	if (strcmp(p->checksuffix, "t") == 0) tagtype = 1013; /* utr/neu */
	else tagtype = 1000; /* jj.pos.utr.sin.ind.nom */
      }
      else if (strcmp(s, "t") == 0 || strcmp(s, "tt") == 0) tagtype = 1001; /* jj.pos.neu.sin.ind.nom */
      else if (IsSuffix(suffix, "are")) tagtype = 1002; /* jj.kom.utr/neu.sin/plu.ind/def.nom */
      else if (strcmp(s, "s") == 0) {
	if (strcmp(p->checksuffix, "t") == 0) tagtype = 1014; /* utr/neu */
	else tagtype = 1003; /* jj.pos.utr.sin.ind.gen */
      }
      else if (strcmp(s, "a") == 0 || strcmp(s, "na") == 0) tagtype = 1004; /* jj.pos.utr/neu.sin.def.nom jj.pos.utr/neu.plu.ind/def.nom */
      else if (strcmp(s, "as") == 0 || strcmp(s, "nas") == 0) tagtype = 1005; /* jj.pos.utr/neu.sin.def.gen jj.pos.utr/neu.plu.ind/def.gen */
      else if (strcmp(s, "e") == 0 || strcmp(s, "ne") == 0) tagtype = 1006; /* jj.pos.mas.sin.def.nom */
      else if (strcmp(s, "es") == 0 || strcmp(s, "nes") == 0) tagtype = 1007; /* jj.pos.mas.sin.def.gen */
      else if (IsSuffix(suffix, "ares")) tagtype = 1008; /* jj.kom.utr/neu.sin/plu.ind/def.gen */
      else if (IsSuffix(suffix, "ts")) tagtype = 1012; /* jj.pos.neu.sin.ind.gen */
    } else tagtype = 0;
  } else
  if (strcmp(wordclass, "vb") == 0) {
    p = suf->check;
    if (p && p->accept) {
      char *s = suffix, *t = p->checksuffix;
      while (*s && *s == *t) { s++; t++; }
      if (IsSuffix(suffix, "ande") || IsSuffix(suffix, "ende"))
	tagtype = 2013; /* pc.prs.utr/neu.sin/plu.ind/def.nom nn.neu.sin.ind.nom */
      else if (IsSuffix(suffix, "andes") || IsSuffix(suffix, "endes"))
	tagtype = 2014; /* pc.prs.utr/neu.sin/plu.ind/def.gen nn.neu.sin.ind.gen ab */
      else if (IsSuffix(suffix, "andet") || IsSuffix(suffix, "endet"))
	tagtype = 2015; /* nn.neu.sin.def.nom */
      else if (IsSuffix(suffix, "andets") || IsSuffix(suffix, "endets"))
	tagtype = 2016; /* nn.neu.sin.def.gen */
      else if (IsSuffix(suffix, "anden") || IsSuffix(suffix, "enden"))
	tagtype = 2017; /* nn.neu.plu.ind.nom */
      else if (IsSuffix(suffix, "andens") || IsSuffix(suffix, "endens"))
	tagtype = 2018; /* nn.neu.plu.ind.gen */
      else if (IsSuffix(suffix, "andena") || IsSuffix(suffix, "endena"))
	tagtype = 2019; /* nn.neu.plu.def.nom */
      else if (IsSuffix(suffix, "andenas") || IsSuffix(suffix, "endenas"))
	tagtype = 2020; /* nn.neu.plu.def.gen */
      else if (strcmp(s, "des") == 0 || strcmp(s, "ddes") == 0 || 
	       strcmp(s, "tes") == 0)
	tagtype = 2006; /* vb.prt.sfo */
      else if (strcmp(s, "tt") == 0)
	tagtype = 2007; /* vb.sup.akt */
      else if (strcmp(s, "tts") == 0)
	tagtype = 2008; /* vb.sup.sfo */
      else if (p->suffixlen == 0 || p->checksuffix[p->suffixlen -1] != 'a') {
	if (!*suffix && !*p->checksuffix) tagtype = 2021; /* vb.inf.akt vb.imp.akt */
	else if (IsSuffix(suffix, "a")) tagtype = 2011;
	else if (IsSuffix(suffix, "as")) {
	  if (strcmp(p->checksuffix, "as") == 0) tagtype = 2022; /* vb.inf.sfo */
	  else tagtype = 2012;
	} 
	else if (strcmp(suffix, "dde") == 0) tagtype = 2002;
	else if (strcmp(suffix, "r") == 0) tagtype = 2003;
	else if (strcmp(suffix, "s") == 0) tagtype = 2023;
	else if (strcmp(suffix, "et") == 0) tagtype = 2026;
	else if (strcmp(suffix, "ets") == 0) tagtype = 2027;
	else if (strcmp(suffix, "ne") == 0) tagtype = 2028;
	else if (strcmp(suffix, "nes") == 0) tagtype = 2029;
	else if (strcmp(suffix, "dd") == 0 || strcmp(suffix, "en") == 0) 
	  tagtype = 2009;
	else if (strcmp(suffix, "dds") == 0 || strcmp(suffix, "ens") == 0) 
	  tagtype = 2010;
	else if (strcmp(suffix, "it") == 0) 
	  tagtype = 2024;
	else if (strcmp(suffix, "its") == 0 || strcmp(suffix, "ats") == 0)
	  tagtype = 2025;
	else tagtype = 0;
      }
      else if (!*s && *t) tagtype = 2000; /* vb.imp.akt */
      else if (!*s) tagtype = 2001; /* vb.inf.akt */
      else if (strcmp(s, "de") == 0 || strcmp(s, "te") == 0 || strcmp(s, "e") == 0)
	tagtype = 2002; /* vb.prt.akt */
      else if (strcmp(s, "er") == 0 || strcmp(s, "r") == 0) 
	tagtype = 2003; /* vb.prs.akt */
      else if (IsSuffix(suffix, "as") && *s != 'd')
	tagtype = 2004; /* vb.inf.sfo */
      else if (strcmp(s, "es") == 0 || strcmp(s, "s") == 0)
	tagtype = 2005; /* vb.prs.sfo */
      else if (strcmp(s, "t") == 0) tagtype = 2007; /* vb.sup.akt */
      else if (strcmp(s, "ts") == 0) tagtype = 2008; /* vb.sup.sfo */
      else if (strcmp(s, "d") == 0) tagtype = 2009; /* pc.prf.utr.sin.ind.nom jj.pos.utr.sin.ind.nom */
      else if (strcmp(s, "ds") == 0) tagtype = 2010; /* pc.prf.utr.sin.ind.gen */
      else if (strcmp(s, "da") == 0 || strcmp(s, "ta") == 0) 
	tagtype = 2011; /* pc.prf.utr.plu.ind.nom */
      else if (strcmp(s, "das") == 0 || strcmp(s, "tas") == 0) 
	tagtype = 2012; /* pc.prf.utr.plu.ind.gen */
      else if (strcmp(s, "dare") == 0) tagtype = 1002; /* jj.kom.utr/neu.sin/plu.ind/def.nom */
      else if (strcmp(s, "dares") == 0) tagtype = 1008; /* jj.kom.utr/neu.sin/plu.ind/def.gen */
      else if (strcmp(s, "dast") == 0) tagtype = 1009; /* jj.suv.utr/neu.sin/plu.ind.nom */
      else if (strcmp(s, "daste") == 0) tagtype = 1010; /* jj.suv.utr/neu.sin/plu.def.nom */
      else if (strcmp(s, "dastes") == 0) tagtype = 1011; /* jj.suv.utr/neu.sin/plu.def.gen */
      else tagtype = 0;
    } else tagtype = 0;
  }

  if (suf->stopsearchatbeginning) printf("<B>");
  if (*fep) {
    printf("(");
    WriteISO(*fep);
    for (fep++; *fep; fep++) {
      printf("|");
      WriteISO(*fep);
    }
    printf(")");
  }
  if (suf->allowedlastcharacters && *suf->allowedlastcharacters) {
    printf("[");
    WriteISO(suf->allowedlastcharacters);
    printf("]");
  }
  if (suf->forbiddenlastcharacters && *suf->forbiddenlastcharacters) {
    unsigned char tmps[MAXLINELEN];
    strcpy(tmps, suf->forbiddenlastcharacters);
    if (IsSuffix(tmps, "sxz") && *suffix != 's')
      tmps[strlen(tmps) - 3] = '\0';
    if (*tmps) {
      printf("[^");
      WriteISO(tmps);
      printf("]");
    }
  }

  if (!suf->accept) /* negerad regel */
    printf("~");
  if (*suffix) WriteISO(suffix);
  else printf("@");
  printf(" \t");
  for (p = suf->check; p; p = p->next) {
    if (p != suf->check) printf(", ");
    if (!p->accept) putchar('~');
    if (*p->checksuffix == '\0') printf("@");
    else {
      WriteISO(p->checksuffix);
    }
  }
  putchar('\t');
  if (!*tag)
  switch (tagtype) {
  case 'n':
  case 't': t = (tagtype == 'n' ? "utr" : "neu");
    switch (linecount) {
    case 1: sprintf(tag, "nn.%s.sin.ind.nom", t); break;
    case 2: if (lastchar != 's') sprintf(tag, "nn.%s.sin.def.nom", t);
            else sprintf(tag, "nn.%s.?.?.gen", t);
            break;
    case 3: if (lastchar != 's') sprintf(tag, "nn.%s.plu.ind.nom", t);
            else {
	      sprintf(tag, "nn.%s.?.?.gen", t);
	      linecount++;
	    }
            break;
    case 4: if (lastchar == 's') sprintf(tag, "nn.%s.sin.ind.gen", t); 
            else sprintf(tag, "nn.%s.?.?.nom", t);
            break;
    case 5: if (lastchar == 's') sprintf(tag, "nn.%s.sin.def.gen", t);
            else sprintf(tag, "nn.%s.?.?.nom", t);
            break;
    case 6: if (lastchar == 's') sprintf(tag, "nn.%s.plu.ind.gen", t);
            else {
	      sprintf(tag, "nn.%s.?.?.nom", t);
	      linecount++;
	    }
            break;
    case 7: if (lastchar != 's') sprintf(tag, "nn.%s.plu.def.nom", t);
            else sprintf(tag, "nn.%s.?.?.gen", t);
            break;
    case 8: if (lastchar == 's') sprintf(tag, "nn.%s.plu.def.gen", t);
            else sprintf(tag, "nn.%s.?.?.nom", t);
            break;
    default: if (lastchar == 's') sprintf(tag, "nn.%s.?.?.gen", t);
            else sprintf(tag, "nn.%s.?.?.nom", t);
            break;
    }
    break;
  case 1000: sprintf(tag, "jj.pos.utr.sin.ind.nom"); break;
  case 1001: sprintf(tag, "jj.pos.neu.sin.ind.nom ab.pos"); break;
  case 1002: sprintf(tag, "jj.kom.utr/neu.sin/plu.ind/def.nom ab.kom"); break;
  case 1003: sprintf(tag, "jj.pos.utr.sin.ind.gen"); break;
  case 1004: sprintf(tag, "jj.pos.utr/neu.sin.def.nom jj.pos.utr/neu.plu.ind/def.nom"); break;
  case 1005: sprintf(tag, "jj.pos.utr/neu.sin.def.gen jj.pos.utr/neu.plu.ind/def.gen"); break;
  case 1006: sprintf(tag, "jj.pos.mas.sin.def.nom"); break;
  case 1007: sprintf(tag, "jj.pos.mas.sin.def.gen"); break;
  case 1008: sprintf(tag, "jj.kom.utr/neu.sin/plu.ind/def.gen"); break;
  case 1009: sprintf(tag, "jj.suv.utr/neu.sin/plu.ind.nom ab.suv"); break;
  case 1010: sprintf(tag, "jj.suv.utr/neu.sin/plu.def.nom"); break;
  case 1011: sprintf(tag, "jj.suv.utr/neu.sin/plu.def.gen"); break;
  case 1012: sprintf(tag, "jj.pos.neu.sin.ind.gen"); break;
  case 1013: sprintf(tag, "jj.pos.utr.sin.ind.nom jj.pos.neu.sin.ind.nom ab"); break;
  case 1014: sprintf(tag, "jj.pos.utr.sin.ind.gen jj.pos.neu.sin.ind.gen"); break;
  case 2000: sprintf(tag, "vb.imp.akt"); break;
  case 2001: sprintf(tag, "vb.inf.akt"); break;
  case 2002: sprintf(tag, "vb.prt.akt pc.prf.mas.sin.def.nom"); break;
  case 2003: sprintf(tag, "vb.prs.akt"); break;
  case 2004: sprintf(tag, "vb.inf.sfo"); break;
  case 2005: sprintf(tag, "vb.prs.sfo"); break;
  case 2006: sprintf(tag, "vb.prt.sfo pc.prf.mas.sin.def.gen"); break;
  case 2007: sprintf(tag, "vb.sup.akt pc.prf.neu.sin.ind.nom ab"); break;
  case 2008: sprintf(tag, "vb.sup.sfo pc.prf.neu.sin.ind.gen"); break;
  case 2009: sprintf(tag, "pc.prf.utr.sin.ind.nom jj.pos.utr.sin.ind.nom"); break;
  case 2010: sprintf(tag, "pc.prf.utr.sin.ind.gen"); break;
  case 2011: sprintf(tag, "pc.prf.utr/neu.sin.def.nom pc.prf.utr/neu.plu.ind/def.nom"); break;
  case 2012: sprintf(tag, "pc.prf.utr/neu.sin.def.gen pc.prf.utr/neu.plu.ind/def.gen"); break;
  case 2013: sprintf(tag, "pc.prs.utr/neu.sin/plu.ind/def.nom nn.neu.sin.ind.nom"); break;
  case 2014: sprintf(tag, "pc.prs.utr/neu.sin/plu.ind/def.gen nn.neu.sin.ind.gen ab"); break;
  case 2015: sprintf(tag, "nn.neu.sin.def.nom"); break;
  case 2016: sprintf(tag, "nn.neu.sin.def.gen"); break;
  case 2017: sprintf(tag, "nn.neu.plu.ind.nom"); break;
  case 2018: sprintf(tag, "nn.neu.plu.ind.gen"); break;
  case 2019: sprintf(tag, "nn.neu.plu.def.nom"); break;
  case 2020: sprintf(tag, "nn.neu.plu.def.gen"); break;
  case 2021: sprintf(tag, "vb.inf.akt vb.imp.akt"); break;
  case 2022: sprintf(tag, "vb.inf.sfo vb.prs.sfo"); break;
  case 2023: sprintf(tag, "vb.inf.sfo vb.prs.sfo"); break;
  case 2024: sprintf(tag, "vb.sup.akt"); break;
  case 2025: sprintf(tag, "vb.sup.sfo"); break;
  case 2026: sprintf(tag, "pc.prf.neu.sin.ind.nom"); break;
  case 2027: sprintf(tag, "pc.prf.neu.sin.ind.gen"); break;
  case 2028: sprintf(tag, "pc.prf.mas.sin.def.nom"); break;
  case 2029: sprintf(tag, "pc.prf.mas.sin.def.gen"); break;

  default: sprintf(tag, "%s", wordclass); /* bara ordklasstagg */

  }
  if (*tag) printf("<%s> ", tag);
}

/* InitSuf initierar suffixtabellen och returnerar 0 om det gick bra. */
int InitSuf(unsigned char *SLfilename)
{ FILE *fp;
  unsigned char line[MAXLINELEN], *s, *start;
  unsigned char tag[MAXLINELEN], *nexttag, ruletype;
  int len, i;
  suffixset *newsuf;
  suffixchecklist *p;
  int tagtype = 0;
  fp = fopen(SLfilename, "r");
  if (!fp) return 1;
  while (fgets(line, MAXLINELEN, fp)) {
    len = strlen(line);
    if (line[len - 1] == '\n') line[--len] = '\0';
    for (i = 0; line[i] == ' '; i++);
    if (len == i) {
      linecount = 0;
      printf("\n");
      continue;
    }
    linecount++;
    s = malloc(len + 1 - i);
    strcpy(s, line + i);
    newsuf = malloc(sizeof(*newsuf));
    newsuf->allowedlastcharacters = newsuf->forbiddenlastcharacters = NULL;
    newsuf->forbiddenending[0] = NULL;
    newsuf->stopsearchatbeginning = 0;
    if (*s == '#' && (s[1] == '#' || s[1] == '*')) {
      ruletype = s[1];
      s += 2; /* ##-mrkta rader ska vara med */
    } else ruletype = '\0';
    if (*s == '<') {
      s++;
      switch (*s) {
      case 'v':
      case 'V': /* Stop search at first vowel (default) */
	newsuf->stopsearchatbeginning = 0;
	break;
      case 'b':
      case 'B': /* Stop search at beginning of word */
	newsuf->stopsearchatbeginning = 1;
	break;
      default:
	fprintf(stderr, "Fel i suffixtabellen. Felaktigt <*>-uttryck.\n");
	continue;
      }
      while (*s && *s++ != '>');
    }
    if (*s == '(') {
      int last = 0, ending = 0;
      unsigned char *t;
      unsigned char **p = newsuf->forbiddenending;
      s++;
      t = s; 
      do {
	for (; *t && *t != ')' && *t != '|'; t++)
	  *t = ISO_intern[*t];
	if (*t) {
	  if (*t == ')') last = 1;
	  *t++ = '\0';
	  newsuf->forbiddenendinglength[ending] = strlen(s);
	  *p++ = s;
	  s = t;
	} else {
	  fprintf(stderr, "Fel i suffixtabellen. Oavslutat (*)-uttryck.\n");
	  break;
	}
      } while (!last && ++ending < MAXNOOFFORBIDDENENDINGS);
      if (!last) {
	if (ending >= MAXNOOFFORBIDDENENDINGS)
	  fprintf(stderr, 
		  "Fel i suffixtabellen. Fr mnga frbjudna ndelser.\n");
	continue; /* oavslutat (*)-uttryck */
      }
      *p = NULL;
    }      
    if (*s == '[') {
      int dis = 0;
      unsigned char *t;
      s++;
      if (*s == '^' || *s == '~') { dis = 1; s++; }
      for (t = s; *t && *t != ']'; t++)
	*t = ISO_intern[*t];
      if (*t) {
	*t++ = '\0';
	if (dis) newsuf->forbiddenlastcharacters = s;
	else newsuf->allowedlastcharacters = s;
	s = t;
      } else {
	fprintf(stderr, "Fel i suffixtabellen. Oavslutat [*]-uttryck.\n");
	continue;
      }
    }      
    if ((s = NextWord(s, &start)) == NULL) {
      if (kommentar && *kommentar) {
	printf("%s", kommentar);
	if (CopyTag(tag, kommentar + 1)) strcpy(wordclass, tag);
	kommentar = NULL;
      }
      printf("\n");
      free(newsuf); 
    }
    else {
      if (*start == '_') { /* negerad regel */
	newsuf->accept = 0;
	start++;
      } else newsuf->accept = 1;
      newsuf->suffix = start;
      newsuf->len = strlen(start);
      newsuf->left = newsuf->right = NULL;
      if ((s = NextWord(s, &start)) == NULL) {
	fprintf(stderr, "Fel i suffixtabellen. Suffixet %s saknar regler.\n",
		newsuf->suffix);
	newsuf->check = NULL;
	continue;
      }
      newsuf->check = p = malloc(sizeof(*p));
      if (*start == '_') { /* negerat suffix */
	p->accept = 0;
	start++;
      } else p->accept = 1;
      p->checksuffix = start;
      while ((s = NextWord(s, &start)) != NULL) {
	p->suffixlen = strlen(p->checksuffix);
	p->next = malloc(sizeof(*p->next));
	p = p->next;
	if (*start == '_') { /* negerat suffix */
	  p->accept = 0;
	  start++;
	} else p->accept = 1;
	p->checksuffix = start;
      }
      p->suffixlen = strlen(p->checksuffix);
      p->next = NULL;
      if (linecount == 1) {
	tagtype = 0;
	p = newsuf->check;
	if (strcmp(wordclass, "nn") == 0) {
	  if (p && p->accept && p->next && p->next->accept) {
	    switch (p->next->checksuffix[p->next->suffixlen - 1]) {
	    case 'n': tagtype = 'n'; break;
	    case 't': tagtype = 't'; break;
	    default:  tagtype = 0;
	    }
	  }
	}

	nexttag = kommentar;
	for (p = newsuf->check; p && p->accept; p = p->next) {
	  nexttag = CopyTag(tag, nexttag);
	  WriteSuffix(newsuf, p->checksuffix, tagtype, tag);
	  printf("#{EL} ");
	  if (kommentar && *kommentar) {
	    WriteWithoutTags(kommentar);
	    kommentar = NULL;
	  }
	  printf("\n");
	  linecount++;
	}
      } else {
	nexttag = kommentar;
      }
      nexttag = CopyTag(tag, nexttag);
      WriteSuffix(newsuf, newsuf->suffix, tagtype, tag);
      if (ruletype == '*') printf("#{EL} "); else
      if (ruletype == '#') printf("#{extra} ");
      if (kommentar && *kommentar) WriteWithoutTags(kommentar);
      printf("\n");
      s = newsuf->forbiddenlastcharacters;
      if (s && IsSuffix(s, "sxz") && strcmp(newsuf->suffix, "s") == 0) {
	*s = '\0'; /* Ta bort sxz */
	newsuf->allowedlastcharacters = "sxz";
	*newsuf->suffix = '\0';
	WriteSuffix(newsuf, newsuf->suffix, tagtype, tag);
	if (ruletype == '*') printf("#{EL} "); else
	  if (ruletype == '#') printf("#{extra} ");
	if (kommentar && *kommentar) WriteWithoutTags(kommentar);
	printf("\n");
      }
    }
  }
  fclose(fp);
  return 0;
}

int main(int argc, char **argv)
{ char *filename;
  *wordclass = '\0';
  InitieraBokstavsTabeller();
  if (argc == 2) filename = argv[1]; else filename = "SLlist";
  fprintf(stderr, "Skapar taggversion av %s\n", filename);
  InitSuf(filename);
  return 0;
}
