/* frekvens.c lser en ordlista med ett ord per rad och berknar bokstavs-
   frekvenser.

   Copyright (C) 1993-2004
   Viggo Kann viggo@nada.kth.se

   Om FYRGRAM definieras produceras ocks en fyrgramsfil som resultat.
   Filen innehller ett booleskt NOOFGRAMS*NOOFGRAMS*NOOFGRAMS*NOOFGRAMS-flt som anger vilka
   kombinationer av fyra bokstver som frekommer i ordlistan.
   A=0, B=1, ..., Z=25, [=26, \=27, ]=28, ~=29, DEL=30.
   DELIMGRAM str fr en ordavskiljare och finns frst och sist i ett ord samt
   vid bindestreck och mellanslag.
   Fltet lagras som en bitvektor med elementet (i,j,k,l) p plats
   ((i * NOOFGRAMS + j) * NOOFGRAMS + k) * NOOFGRAMS + l.

   1998-04-09 (konstanterna NOOFGRAMS, DELIMGRAM infrda)
   2001-09-27 intern_gram anvnds
*/

/******************************************************************************

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define FYRGRAM
/* #define ENGELSKA */

#ifdef FYRGRAM
#include "stava.h"
#include "stavaconstants.h"
#define FYRGRAMFIL "fyrgraf"
#endif /* FYRGRAM */

int main(void)
{ unsigned char buf[305], *s;
  long f[256], rad = 0L;
  int i, j, len;
#ifdef FYRGRAM
  int k;
  FILE *fp;
  unsigned char fyrtabell[FGRAMSIZE];
  long fyrgram[NOOFGRAMS][NOOFGRAMS][NOOFGRAMS][NOOFGRAMS], anvanda = 0, l;

  for (l = 0; l < FGRAMSIZE; l++) fyrtabell[l] = 0;
  for (i = 0; i < NOOFGRAMS; i++)
    for (j = 0; j < NOOFGRAMS; j++)
     for (k = 0; k < NOOFGRAMS; k++)
      for (len = 0; len < NOOFGRAMS; len++)
	fyrgram[i][j][k][len] = 0;
#endif /* FYRGRAM */
  for (i = 0; i < 256; i++) f[i] = 0;
  buf[300] = '\0';
  while (fgets((char *) buf + 1, 299, stdin)) {
    buf[0] = '-';
    rad++;
    len = strlen((char *) buf) - 1;
    if (buf[len] == '\n') len--;
    buf[len+1] = '\0';
    for (s = buf + 1; *s; s++) f[*s]++;
#ifdef FYRGRAM
    buf[len+1] = '-';
    buf[len+2] = '\0';
    if (len >= 2) {
      for (s = buf; *s; s++) *s = intern_gram[*s];
      *s = 255;
      for (s = buf + 3; *s != 255; s++) {
/*	if (*(s-3) == 4 && *(s-2) == 15 && *(s-1) == 19 && *s == DELIMGRAM)
	  printf("innehller ept~\n"); */
	fyrgram[*(s-3)][*(s-2)][*(s-1)][*s]++;
      }
    }
#endif /* FYRGRAM */
  }
  fprintf(stderr, "\n");
#ifdef FYRGRAM
  for (i = 0; i < NOOFGRAMS; i++)
    for (j = 0; j < NOOFGRAMS; j++)
     for (k = 0; k < NOOFGRAMS; k++)
      for (len = 0; len < NOOFGRAMS; len++) {
/*	if (i == 4 && j == 15 && k == 19 && len == DELIMGRAM)
	  printf("Sammanlagt %d ept-frekoster\n", fyrgram[i][j][k][len]); */
	if (fyrgram[i][j][k][len]) {
	  anvanda++;
	  l = (((long) i * NOOFGRAMS + j) * NOOFGRAMS + k) * NOOFGRAMS + len;
	  fyrtabell[l >> 3] |= (1 << ((int) l & 7));
	}
      }

/*  l = ((4L * NOOFGRAMS + 15) * NOOFGRAMS + 19) * NOOFGRAMS + DELIMGRAM;
  printf("Svar: fyrtabell[%d] & (1 << %d) = %ld\n", l >> 3, (int)(l & 7),
	 (long) (fyrtabell[l >> 3] & (1 << ((int) l & 7)))); */
  printf("%ld fyrgram anvndes, dvs %5.2f procent.\n", anvanda,
	 100.0 * (float) anvanda / (NOOFGRAMS*NOOFGRAMS*NOOFGRAMS*NOOFGRAMS));
  anvanda = 0;
  for (i = 0; i < NOOFGRAMS; i++)
    for (j = 0; j < NOOFGRAMS; j++)
     for (k = 0; k < NOOFGRAMS; k++)
      for (len = 0; len < NOOFGRAMS; len++)
	if (fyrgram[i][j][k][len] > 1) anvanda++;
  printf("%ld fyrgram anvndes mer n en gng, dvs %5.2f procent.\n",
	 anvanda, 100.0 * (float) anvanda / (NOOFGRAMS*NOOFGRAMS*NOOFGRAMS*NOOFGRAMS));
#endif /* FYRGRAM */
  printf("Sammanlagt %ld rader. Frekvenstabell:", rad);
  j = 0;
  for (i = 0; i < 256; i++)
    if (f[i]) {
      if (j++ % 6 == 0) printf("\n");
      if ((i >= ' ' && i < 127) || i >= 160) printf("%c%9ld   ", i, f[i]);
      else printf("%3d%7ld   ", i, f[i]);
    }
  printf("\n");
#ifdef FYRGRAM
  if (!(fp = fopen(FYRGRAMFIL, "wb"))) {
    fprintf(stderr, "Kan inte ppna %s\n", FYRGRAMFIL);
    exit(1);
  }
  printf("Fyrgramstabellen skrivs p filen %s.\n", FYRGRAMFIL);
  fwrite(fyrtabell, sizeof(unsigned char), FGRAMSIZE, fp);
  fclose(fp);
#endif /* FYRGRAM */
  return 0;
}
