/* Rttstavningsprogram. Version 2.21  1997-09-29
   Copyright (C) 1990-1997
   Joachim Hollman och Viggo Kann
   joachim@nada.kth.se viggo@nada.kth.se
*/

/******************************************************************************

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

******************************************************************************/

/* Definiera en av ELCREATE, FLCREATE, ILCREATE, ULCREATE eller XLCREATE */
/* #define ELCREATE */
/* #define FLCREATE */
/* #define ILCREATE */
/* #define ULCREATE */
/* #define XLCREATE */

/* #define ENGELSKA */ /* Engelsk stavningskontroll */
#define SVENSKA /* Svensk stavningskontroll */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "stava.h"

#include "stavaconstants.h"

#ifdef ELCREATE
#define SIZE        ELSIZE
#define NOOFHASH    ELNOOFHASH
#define FILENAME    ELFILENAME
#define BITS       ELBITS
#endif
#ifdef FLCREATE
#define SIZE        FLSIZE
#define NOOFHASH    FLNOOFHASH
#define FILENAME    FLFILENAME
#define BITS       FLBITS
#endif
#ifdef ILCREATE
#define SIZE        ILSIZE
#define NOOFHASH    ILNOOFHASH
#define FILENAME    ILFILENAME
#define BITS       ILBITS
#endif
#ifdef ULCREATE
#define SIZE        ULSIZE
#define NOOFHASH    ULNOOFHASH
#define FILENAME    ULFILENAME
#define BITS       ULBITS
#endif
#ifdef XLCREATE
#include "rattstava.h"
#define SIZE        XLSIZE
#define NOOFHASH    XLNOOFHASH
#define FILENAME    XLFILENAME
#define BITS       XLBITS
#endif

unsigned char HUGEVAR table[SIZE];
unsigned char *bokstavsTabell;
FILE *fp;

static INLINE long myfwrite(void *p, size_t elsize, long n, FILE *fp)
{ return fwrite(p, elsize, n, fp); }

static long antalEttor(unsigned char HUGEVAR table[], long size)
{ long antal = 0L, i;
  int a;
  for (i = 0; i < size; i++) {
    a = table[i];
    while (a) {
      if (a & 1) antal++;
      a >>= 1;
    }
  }
  printf("%ld ettor av %ld mjliga, dvs %f\n", antal, 8 * size,
	 (float) antal / (8 * size));
  printf("Felsannolikhet: %.2e\n",
	 pow((double) antal / (8 * size), (double) NOOFHASH));
  return antal;
}

typedef  unsigned long  int  ub4;   /* unsigned 4-byte quantities */

#define hashsize(n) ((ub4)1<<(n))
#define hashmask(n) (hashsize(n)-1)


/*
--------------------------------------------------------------------
mix -- mix 3 32-bit values reversibly.
For every delta with one or two bit set, and the deltas of all three
  high bits or all three low bits, whether the original value of a,b,c
  is almost all zero or is uniformly distributed,
* If mix() is run forward or backward, at least 32 bits in a,b,c
  have at least 1/4 probability of changing.
* If mix() is run forward, every bit of c will change between 1/3 and
  2/3 of the time.  (Well, 22/100 and 78/100 for some 2-bit deltas.)
mix() takes 36 machine instructions, but only 18 cycles on a superscalar
  machine (like a Pentium or a Sparc).  No faster mixer seems to work,
  that's the result of my brute-force search.  There were about 2^^68
  hashes to choose from.  I only tested about a billion of those.
--------------------------------------------------------------------
*/
#define mix(a,b,c) \
{ \
  a -= b; a -= c; a ^= (c>>13); \
  b -= c; b -= a; b ^= (a<<8); \
  c -= a; c -= b; c ^= (b>>13); \
  a -= b; a -= c; a ^= (c>>12);  \
  b -= c; b -= a; b ^= (a<<16); \
  c -= a; c -= b; c ^= (b>>5); \
  a -= b; a -= c; a ^= (c>>3);  \
  b -= c; b -= a; b ^= (a<<10); \
  c -= a; c -= b; c ^= (b>>15); \
}

static INLINE void Store(unsigned char *k)
{
   register ub4 a,b,c;
   register ub4 h, olda, oldb;
   register int i;
   register int len = strlen((char *)k), length = len;

   /* Set up the internal state */
   a = b = 0x9e3779b9;  /* the golden ratio; an arbitrary value */
   c = HASHINITVAL;           /* the previous hash value */

   /*---------------------------------------- handle most of the key */
   while (len >= 12)
   {
      a += (k[0] +((ub4)k[1]<<8) +((ub4)k[2]<<16) +((ub4)k[3]<<24));
      b += (k[4] +((ub4)k[5]<<8) +((ub4)k[6]<<16) +((ub4)k[7]<<24));
      c += (k[8] +((ub4)k[9]<<8) +((ub4)k[10]<<16)+((ub4)k[11]<<24));
      mix(a,b,c);
      k += 12; len -= 12;
   }

   /*------------------------------------- handle the last 11 bytes */
   c += length;
   switch(len)              /* all the case statements fall through */
   {
   case 11: c+=((ub4)k[10]<<24);
   case 10: c+=((ub4)k[9]<<16);
   case 9 : c+=((ub4)k[8]<<8);
      /* the first byte of c is reserved for the length */
   case 8 : b+=((ub4)k[7]<<24);
   case 7 : b+=((ub4)k[6]<<16);
   case 6 : b+=((ub4)k[5]<<8);
   case 5 : b+=k[4];
   case 4 : a+=((ub4)k[3]<<24);
   case 3 : a+=((ub4)k[2]<<16);
   case 2 : a+=((ub4)k[1]<<8);
   case 1 : a+=k[0];
     /* case 0: nothing left to add */
   }
   olda = a; oldb = b;
   mix(a,b,c);
   /*-------------------------------------------- report the result */
   h = c & hashmask(BITS);
   table[h >> 3] |= (1 << (int)(h & 7));
   for (i = 1; i < NOOFHASH; i++) {
     a = olda; b = oldb;
     mix(a,b,c);
     h = c & hashmask(BITS);
     table[h >> 3] |= (1 << (int)(h & 7));
   }
}


static void InitieraBokstavsTabeller(void)
{ int i;
  bokstavsTabell = ISO_intern;
/* bokstavsTabell = ASCII_intern; */
  /* fyll i 0-flt med standardvrde i bokstavstabellen till ISO */
  for (i = 0; i < 256; i++) {
    if (intern_ISO[i] == 0) intern_ISO[i] = i;
  }
}

static int TagOrd(unsigned char *word, unsigned char *urord)
{ int len;
  unsigned char *s, *t;
  unsigned char buf[1000];
 start:
  if (!gets((char *) buf)) return 0;
  for (len = strlen((const char *) buf); len > 0 &&
	 (buf[len-1] == '\r' || buf[len-1] == ' '); len--)
    buf[len-1] = '\0';
  for (s = buf; *s == ' '; s++);
  if (!*s) goto start;
  strcpy((char *) urord, (const char *) buf);
  for (s = urord, t = word; *s; s++) {
    *t = bokstavsTabell[*s];
    if (!*t) {
      fprintf(stderr, "Felaktigt ord: %s.\n", urord);
      goto start;
    }
    t++;
  }
  *t = '\0';
  return 1;
}

static void Initiera(void)
{
  InitieraBokstavsTabeller();
}

static void OppnaFiler(void)
{
  /* ppna lexikonfil: */
  if (!(fp = fopen(LIBPATH FILENAME,"wb"))) {
    fprintf(stderr,"Kan inte ppna filen %s\n", LIBPATH FILENAME);
    exit(1);
  }
  antalEttor(table, SIZE);
  myfwrite(table, sizeof(unsigned char), SIZE, fp);
  fclose(fp);
}

int main(void)
{ unsigned char ordin[LANGD + 4], urord[LANGD + 200];
  long antalOrd = 0;

#ifdef ELCREATE
  fprintf(stderr, "Detta program skapar en ny EL-ordlista %s.\n",
      ELFILENAME);
#endif
#ifdef FLCREATE
  fprintf(stderr, "Detta program skapar en ny FL-ordlista %s.\n",
      FLFILENAME);
#endif
#ifdef ILCREATE
  fprintf(stderr, "Detta program skapar en ny IL-ordlista %s.\n",
      ILFILENAME);
#endif
#ifdef ULCREATE
  fprintf(stderr, "Detta program skapar en ny UL-ordlista %s.\n",
      ULFILENAME);
#endif
#ifdef XLCREATE
  fprintf(stderr, "Detta program skapar en ny frekvens-ordlista %s.\n",
      XLFILENAME);
#endif

  Initiera();
  while (TagOrd(ordin, urord)) {
    antalOrd++;
    Store(ordin);
  }
  OppnaFiler();
  fprintf(stderr, "Antal ord: %ld\n", antalOrd);
  return 0;
}
