/* ---------------------- for sockets ---------------------- */
/* add these compiler options or things will not work correctly: */
/* -DAPI -DRATTSTAVA */
/* You also need to change the hard coded libpath in the Makefile */

#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>
#include <strings.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include "stavaapi.h"

// Assuming the text is Swedish text, this should return 1 if the
// encoding is UTF-8, and it should return 0 if we find something that
// looks like latin-1
int looksLikeUTF(const char *s) {
  if(s == NULL) {
    return 0;
  }

  int last = 1;
  for(int i = 0; s[i] != '\0'; i++) {
    if(s[i] < 0 && last > 0) {
      if(s[i] == -27
	 || s[i] == -28
	 || s[i] == -10
	 || s[i] == -59
	 || s[i] == -60
	 || s[i] == -42) {
	return 0;
      }
    }
    if(s[i] < 0 && last == -61) {
      return 1;
    }
    last = s[i];
  }
  
  return 0;
}

/* Converts UTF-8 to Latin-1 correctly if we assume the string
   contents are Swedish text. Based on code example from Mark Ransom
   
   returns new size of buffer */
int utf2latin1(char *s, int len) {
  
  int n = len;
  char *tempBuf = malloc((n+1) * sizeof(char));

  unsigned int codepoint = 0;

  int j = 0;
  for (int i = 0; s[i] != 0; ) {
    unsigned char c = (unsigned char) (s[i]);
    if (c <= 0x7f)
      codepoint = c;
    else if (c <= 0xbf)
      codepoint = (codepoint << 6) | (c & 0x3f);
    else if (c <= 0xdf)
      codepoint = c & 0x1f;
    else if (c <= 0xef)
      codepoint = c & 0x0f;
    else
      codepoint = c & 0x07;

    i++;
    
    if (((s[i] & 0xc0) != 0x80) && (codepoint <= 0x10ffff)) {
      if (codepoint <= 255) {
	tempBuf[j++] = (char)(codepoint);
      } else {
	// do whatever you want for out-of-bounds characters
	tempBuf[j++] = 'X'; // Store something away so we can convert back later? TODO: Jonas
      }
    }
  }

  for(int i = 0; i < j; i++) {
    s[i] = tempBuf[i];
  }
  s[j] = 0;

  free(tempBuf);

  return j;
}

const int SOCKETBUFSIZE = 1000000;
int read_data(int socket, unsigned char *buf) {
  /* char buf[SOCKETBUFSIZE+1]; */
  int resultCount = 0;
  int done = 0;
  int len = 0, offs = 0;
  unsigned char *end = 0;
  int i = 0;
  int lastWasDelim = 1;
  
  buf[0] = 0;
  buf[SOCKETBUFSIZE] = 0;

  while(offs < SOCKETBUFSIZE && !done) {
    len = read(socket, buf+offs, SOCKETBUFSIZE - offs);
    if(len > 0) {
      buf[SOCKETBUFSIZE] = 0;

      end = (unsigned char *)strstr((char *)buf, "ENDQ");
      if(end) {
	*end = 0;
	done = 1;
      } else {
	offs += len;
      }
      
    } else {
      buf[offs] = 0;
      end = buf + offs;
      done = 1;
    }

    if(offs >= SOCKETBUFSIZE) {
      buf[SOCKETBUFSIZE-1] = 0;
      end = &(buf[SOCKETBUFSIZE-1]);
      done = 1;
    }
  }

  if(looksLikeUTF((char *)buf)) {
    int newSize = utf2latin1((char *)buf, end - buf);
    end = buf + newSize;
  }
  
  for( ; i < SOCKETBUFSIZE && i < end - buf; i++) {
    if(buf[i] == 0
       || buf[i] == ' '
       || buf[i] == '\n'
       || buf[i] == '\r'
       || buf[i] == '\t'
       ) {
      if(!lastWasDelim) {
	resultCount++;
      }
      lastWasDelim = 1;
      buf[i] = 0;
    } else {
      lastWasDelim = 0;
    }
  }
  return resultCount;
}

// taken from a socket primer
int establish(int portnum)
{ 
  const int MAXHOSTNAME = 1000;
  char myname[MAXHOSTNAME+1];
  int s;
  struct sockaddr_in sa;
  struct hostent *hp;
  
  bzero(&sa,sizeof(struct sockaddr_in));
  gethostname(myname,MAXHOSTNAME);
  hp= gethostbyname(myname);
  if (hp == NULL)
    return -1;
  sa.sin_family= hp->h_addrtype; /* this is our host address */
  sa.sin_port= htons(portnum); /* this is our port number */
  if ((s = socket(AF_INET,SOCK_STREAM,0)) < 0)
    return -1;
  if (bind(s,(struct sockaddr *)&sa,sizeof(sa)) < 0) {
    close(s);
    return -1;
  }
  listen(s, 5); /* max # of queued connects */
  return s;
}

// taken from a socket primer
int get_connection(int s // socket
		   )
{ 
  struct sockaddr_in isa;
  socklen_t i;
  int t;
  
  i = sizeof(isa);
  getsockname(s,(struct sockaddr *)&isa,&i);
  
  if ((t = accept(s,(struct sockaddr *)&isa,&i)) < 0)   /* accept connection if there is one */
    return -1;
  return t;
}
/* ---------------------- for sockets ---------------------- */

int main(int argc, char **argv)
{
  int idx = 0;
  int portno = 0;
  int server_socket = 0;
  int socketOK = 0;
  int client_socket = 0;
  unsigned char socketBuf[SOCKETBUFSIZE+1];
  int noofwords = 0;
  const unsigned char *corrs = 0;
  int corrlen = 0;
  int inside = 0;
  const char *libpath;
  
  /* handler for child processes */
  signal(SIGCHLD, SIG_IGN);

  for(idx = 0; idx < argc; idx++) {
    if(argv[idx][0] == '-'
       && argv[idx][1] == 'p'
       && argv[idx][2] == 'n') {
      
      portno = atoi(argv[idx] + 3);

      if ((server_socket = establish(portno)) < 0) {  /* plug in the phone */
	fprintf(stderr, "%s: could not open socket for listening\n", argv[0]);
	exit(1);
      } else {
	socketOK = 1;
      }
    }
  }
  if(!socketOK) {
    printf("please specify a port number to listen to with %s -pnXXXX, for example %s -pn6145\n", argv[0], argv[0]);
    exit(0);
  }

  if (!(libpath = getenv("STAVALIB"))) {
    libpath = LIBPATH;
  }
  if(!StavaReadLexicon(libpath, 1, 1, 1, 1, 1, 1, (const unsigned char *)" ")) {
    fprintf(stderr, "Could not read lexicon files correctly, aborting.\n");
    exit(1);
  }
  

  while(1) {
    fprintf(stderr, "%s: waiting for input on port %d\n", argv[0], portno);
    client_socket = 0;
    if ((client_socket = get_connection(server_socket)) < 0) {
      if (errno == EINTR)
	continue;
      fprintf(stderr, "%s: error on accept\n", argv[0]);
      exit(1);
    }
    
    // fork all client calls, that way, if they hang or crash,
    // the server keeps working
    
    switch(fork()) {
    case -1 : // error
      fprintf(stderr, "%s: error on fork()\n", argv[0]);
      exit(1);
    case 0 : // child
      close(server_socket);
      alarm(60*2); // if we still haven't finished in 2 minutes, die...

      noofwords = read_data(client_socket, socketBuf);
      alarm(0); // remove alarm

      inside = 0;
      for(idx = 0; noofwords > 0 && idx < SOCKETBUFSIZE; idx++) {
	if(socketBuf[idx] == 0) {
	  if(inside) {
	    noofwords--;
	  }
	  inside = 0;
	} else {
	  if(!inside) {
	    unsigned char wordBuf[4096+1];
	    int stavaRes = 0;
	    
	    StavaStringGetWord(socketBuf + idx, wordBuf);
	    if (!StavaWord(wordBuf)) {
	      corrs = StavaCorrectWord(wordBuf);
	      if (corrs) {
		corrlen = strlen((char *)wordBuf);
		write(client_socket, wordBuf, corrlen * sizeof(char));
		write(client_socket, ": ", 2 * sizeof(char));
		corrlen = strlen((char *)corrs);
		write(client_socket, corrs, corrlen * sizeof(char));
		write(client_socket, "\n", 1 * sizeof(char));
	      } else { /* word is spelled correctly */
		corrlen = strlen((char *)wordBuf);
		write(client_socket, wordBuf, corrlen * sizeof(char));
		write(client_socket, ": ", 2 * sizeof(char));
		write(client_socket, wordBuf, corrlen * sizeof(char));
		write(client_socket, "\n", 1 * sizeof(char));
	      }
	    } else { /* word is spelled correctly */
	      corrlen = strlen((char *)wordBuf);
	      write(client_socket, wordBuf, corrlen * sizeof(char));
	      write(client_socket, ": ", 2 * sizeof(char));
	      write(client_socket, wordBuf, corrlen * sizeof(char));
	      write(client_socket, "\n", 1 * sizeof(char));
	    }
	  }
	  inside = 1;
	}
      }
      
      close(client_socket);
      exit(0); // child is finished
    default : // parent
      close(client_socket);
      wait(NULL); // wait for child to finish, so as not to spam too many processes
    }
  }
  
  return 0;
}
